/*
 * spin.h  --  Model checker core header file
 *
 * NOTES:
 * 
 *  This file is part of the dSpin3.2 distribution.
 *  Original code written by Gerard Holzmann <gerard@research.bell-labs.com>.
 *  dSpin extension written by Radu Iosif <iosif@athena.polito.it>.
 */


#ifndef __SPIN_H__
#define __SPIN_H__

#include <stddef.h>
#include <errno.h>
#include <sys/time.h>

/*** Some macro dependencies ***/

#ifdef  SAFETY
#ifndef NOFAIR
#define NOFAIR
#endif  /* NOFAIR */
#endif  /* SAFETY */

#ifdef  NOREDUCE
#define XUSAFE
#if     !defined(SAFETY) && !defined(MA)
#define FULLSTACK
#endif  /* !defined(SAFETY) && !defined(MA) */
#else   /* NOREDUCE */
#ifdef  BITSTATE
#ifdef  SAFETY
#define CNTRSTACK
#else   /* SAFETY */
#define FULLSTACK
#endif  /* SAFETY */
#else   /* BITSTATE */
#define FULLSTACK
#endif  /* BITSTATE */
#endif  /* NOREDUCE */

#ifdef  BITSTATE
#define NOCOMP
#endif  /* BITSTATE */

#ifdef  VERBOSE
#define CHECK
#define DEBUG
#endif  /* VERBOSE */

#define DELTA	500

#ifdef   MA
#if     (MA == 1)
#undef   MA
#define  MA	100
#endif
#endif /* MA */

#ifdef  W_XPT
#if    (W_XPT == 1)
#undef  W_XPT
#define W_XPT 1000000
#endif
#endif /* W_XPT */

/* Accept the above for backward compatibility */
#if defined(COLLAPSE2) || defined(COLLAPSE3) || defined(COLLAPSE4)
#define COLLAPSE
#endif

#ifdef COLLAPSE
static unsigned long ncomps [256 + 2];
#endif

#ifndef MEMCNT
#ifdef  PC
#define MEMCNT	25	/* 32 Mb */
#else
#define MEMCNT	28
#endif
#endif  /* MEMCNT */

#ifdef VERI
#define BASE	1
#else
#define BASE	0
#endif /* VERI */

#if SYNC
#define IfNotBlocked	if (boq != -1) continue;
#define UnBlock     	boq = -1
#else
#define IfNotBlocked	/* cannot block */
#define UnBlock     	/* don't bother */
#endif /* SYNC */

#ifndef NOBOUNDCHECK
#define Index(x, y)	Boundcheck(x, y, II, tt, t)
#else
#define Index(x, y)	x
#endif /* NOBOUNDCHECK */

#ifdef CNTRSTACK
#define _onstack_now()	(LL[trpt->j6] && LL[trpt->j7])
#define _onstack_put()	 LL[trpt->j6]++; LL[trpt->j7]++
#define _onstack_zap()	 LL[trpt->j6]--; LL[trpt->j7]--
#endif /* CNTRSTACK */

#ifdef MA
#define _onstack_put()	;
#define _onstack_zap()	_gstore(&now, 4)
#define INLINE_REV
unsigned int maxgs = 0;
#else /* MA */
#if defined(FULLSTACK) && !defined(BITSTATE)
#define _onstack_put()	trpt->ostate = Lstate
#define _onstack_zap()	do {                            \
                           if (trpt->ostate)            \
			     trpt->ostate->tagged = 0;  \
			} while (0)
#endif /* FULLSTACK && BITSTATE */
#endif /* MA */


/*** Stack frame layout -- [ot | tt | bp > ***/

#ifdef __ALIGN
#define framesz                                  \
(sizeof(bp_t) + sizeof(tt_t) + sizeof(ot_t) +    \
PAD(sizeof(bp_t) + sizeof(tt_t) + sizeof(ot_t)))
#else
#define framesz                                  \
(sizeof(bp_t) + sizeof(tt_t) + sizeof(ot_t))      
#endif /* __ALIGN */

#define __bp(__pp)                    \
(*((bp_t *)((__pp)->stk + (__pp)->_sp \
	    - sizeof(bp_t))))

#define getbp(__pp, __bp)              \
(*((bp_t *)((__pp)->stk + (__bp)       \
            - sizeof(bp_t))))

#define __tt(__pp)                    \
(*((tt_t *)((__pp)->stk + (__pp)->_sp \
	    - sizeof(bp_t) - sizeof(tt_t))))

#define gettt(__pp, __bp)             \
(*((ot_t *)((__pp)->stk + (__bp)      \
            - sizeof(bp_t) - sizeof(tt_t))))

#define __ot(__pp)                    \
(*((ot_t *)((__pp)->stk + (__pp)->_sp \
	    - sizeof(bp_t) - sizeof(tt_t) - sizeof(ot_t))))

#define getot(__pp, __bp)             \
(*((ot_t *)((__pp)->stk + (__bp)      \
            - sizeof(bp_t) - sizeof(tt_t) - sizeof(ot_t))))


/*** Variables ***/

#define bset(__n, __p, __v)                          \
do {                                                 \
  uchar b = (__v) ? 1 : 0;                           \
  (__n) = ((__n) | (b << (__p))) & (~(!b << (__p))); \
} while (0)

#define bget(__n, __p)        (((__n) & (1 << (__p))) >> (__p))

#define gg(__off, __t)        (*((__t *)(now.heap + (__off))))
#define gptr(__off)           ((uint)((char *)now.heap + (__off)))

#define ll(__off, __t)        (*((__t *)(this->stk + this->_bp + (__off))))
#define lptr(__off)           ((uint)((char *)this->stk + this->_bp + (__off)))

#define getll(__pp, __off, __bp, __t) \
(*((__t *)((__pp)->stk + (__bp) + (__off))))


/*** Pointers ***/

#define xx                    xptr
#define mp                    mkptr
#define dp                    deptr

#define NPTR                  0
#define LPTR                  1
#define GPTR                  2
#define HPTR                  3
#define DPTR                 -1

uint null = 0;


/*** Other macros ***/

#define efd stderr
#define max(a, b)             (((a) >= (b)) ? (a) : (b))

#ifndef HEAPSZ
#define HEAPSZ  1024
#endif

#define MAXPROC 255
#define MAXQ    255
#define MAXHEAP 255

#ifndef NFAIR
#define NFAIR   2
#endif

#define INLINE	1

P *noptr;  /* used by macro Pptr(x) */
P *this;

#define pptr(x)         now.pool[x]
#define p_sz(x)         ((long)(offsetof(P, stk) + (x)->_sp))


/*** Scalar value ***/

#ifdef POSYMM
typedef struct {
  int len;
  uchar *word;
} scalar_t;
#endif /* POSYMM */


/*** Global queue descriptor ***/

typedef struct {
#if HEAPSZ < 32768
  short _offset;
  short _size;
#else
  long _offset;
  long _size;
#endif 
  short _proc;

#ifdef POSYMM
  scalar_t _scalar;
#endif /* POSYMM */
} q_desc;

q_desc q_tab[MAXQ];

#define q_offset(x)     q_tab[x]._offset
#define q_size(x)       q_tab[x]._size
#define q_proc(x)       q_tab[x]._proc
#define q_scalar(x)     q_tab[x]._scalar
#define q_free(x)       (q_tab[x]._offset == -1)

#define qptr(x)  \
((Q *)(((q_proc(x) == -1) ? now.heap : pptr(q_proc(x))->stk) + q_offset(x)))

#define Q_GLOBAL    0
#define Q_HEAP      1
#define Q_LOCAL     2


/*** Global heap object descriptor ***/

typedef struct {
#if HEAPSZ < 32768
  short _offset;
  short _size;
  short _range;
#else
  long _offset;
  long _size;
  long _range;
#endif

  uint _type;

#ifdef GC0
  ushort _ref;
#endif /* GC0 */

#ifdef POSYMM
  scalar_t _scalar;
#endif /* POSYMM */  
} h_desc;

h_desc h_tab[MAXHEAP];

#define h_offset(x)     h_tab[x]._offset
#define h_size(x)       h_tab[x]._size
#define h_range(x)      h_tab[x]._range
#define h_type(x)       h_tab[x]._type
#define h_ref(x)        h_tab[x]._ref
#define h_scalar(x)     h_tab[x]._scalar
#define h_free(x)       (h_tab[x]._offset == -1)


#ifdef GC1
int h_ordinal = 0;
int h_mark[MAXHEAP];
#endif /* GC1 */


#define LOCAL			1
#define Q_FULL_F		2
#define Q_EMPT_F		3
#define Q_EMPT_T		4
#define Q_FULL_T		5
#define TIMEOUT_F		6
#define GLOBAL			7
#define BAD			8
#define ALPHA_F			9


/*** State type ***/

typedef struct State {
  uchar _nr_pr;
  uchar _nr_qs;
  uchar _a_t;          /* cycle detection */

#ifndef NOFAIR
  uchar _cnt[NFAIR];   /* counters, weak fairness */
#endif

#ifdef HAS_LAST
  uchar _last;
#endif

#ifdef EVENT_TRACE
#if nstates_event < 256
  uchar _event;
#else
  unsigned short _event;
#endif
#endif

#if HEAPSZ < 65536
  unsigned short _hsz;
#else
  unsigned long _hsz;
#endif  

  uchar *heap;
  P    **pool;
} State;

State	A_Root;		/* root of acceptance cycles */
State	now;		/* the full state structure */

int hmax = 0, svmax = 0, smax = 0;

#define POOLSZ     (sizeof(State) + HEAPSZ + MAXPROC * (sizeof(P) * STACKSZ))


/*** Stack for deleted processes and queues ***/

typedef struct Stack {
#if SYNC
  short o_boq;
#endif

  int o_proc;
  int o_offset;
  int o_size;
  int o_delqs;

  uchar o_qs[MAXQ];

#ifndef XUSAFE
  char *o_name;
#endif

  union {
    P *process;
    Q *queue;
  } body;

  struct Stack *nxt;
  struct Stack *lst;
} Stack;

Stack	* stack;

#define S_P(x)  ((x) -> body.process)
#define S_Q(x)  ((x) -> body.queue)


/*** Stack for full state vectors ***/

typedef struct Svtack {
#if SYNC
  short o_boq;
#endif

  q_desc *q_tab[MAXQ];
  h_desc *h_tab[MAXHEAP];

  State *body;

  struct Svtack *nxt;
  struct Svtack *lst;
} Svtack;

Svtack	*svtack;


/*** Transition table ***/
  
typedef struct Trans {
  short atom;	             /* if &2 = atomic trans; if &8 local */
  short st;	             /* the nextstate */
#ifdef HAS_UNLESS
  short escp[HAS_UNLESS];    /* lists the escape states */
  short e_trans;	     /* if set, this is an escp-trans */
#endif
  short tpe[2];	             /* class of operation (for reduction) */
  short qu[6];	             /* for conditional selections: qid's  */
  uchar ty[6];	             /* ditto: type's */
#ifdef NIBIS
  short om;	             /* completion status of preselects */
#endif
  char *tp;	             /* src txt of statement */
  int t_id;	             /* transition id, unique within proc */
  int forw;	             /* index forward transition */
  int back;	             /* index return  transition */
  struct Trans *nxt;
} Trans;

Trans ***trans;	


/*** Hash table ***/

struct H_el {
  struct H_el *nxt;

#ifdef FULLSTACK
  unsigned tagged;
#if defined(BITSTATE) && !defined(NOREDUCE) && !defined(SAFETY)
  unsigned proviso;
#endif
#endif /* FULLSTACK */

#if defined(CHECK) || (defined(COLLAPSE) && !defined(FULLSTACK))
  unsigned long st_id;
#endif

#ifdef COLLAPSE
  unsigned long ln;
#endif /* COLLAPSE */

#ifdef REACH
  unsigned D;
#endif

  unsigned state;
} **H_tab, **S_tab;

/* State space size */
int ssize;

#if     defined(FULLSTACK)
struct H_el *Lstate;
#endif  /* defined(FULLSTACK) */


/*** Trail ***/

typedef struct Trail {
  short st;	/* current state */
  uchar pr;	/* process id */
  uchar tau;	/* 8 different bit-flags */
  uchar o_pm;	/* a few more bit-flags */

#if 0
  Meaning of bit-flags:
  tau&1    -> timeout enabled
  tau&2    -> request to enable timeout 1 level up (in claim)
  tau&4    -> current transition is a  claim move
  tau&8    -> current transition is an atomic move
  tau&16   -> last move was truncated on stack
  tau&32   -> current transition is a preselected move
  tau&64   -> at least one next state is not on the stack
  tau&128  -> current transition is a stutter move
  o_pm&1   -> the current pid moved -- implements else
  o_pm&2   -> this is an acceptance state
  o_pm&4   -> this is a  progress state
  o_pm&8   -> fairness alg rule 1 undo mark
  o_pm&16  -> fairness alg rule 3 undo mark
  o_pm&32  -> fairness alg rule 2 undo mark
  o_pm&64  -> the current proc applied rule2
  o_pm&128 -> a fairness, dummy move - all procs blocked
#endif

#if defined(FULLSTACK) && defined(MA)
  uchar proviso;
#endif
  char  o_n, o_ot, o_m;	/* to save locals */

#ifdef EVENT_TRACE
#if nstates_event < 256
  uchar o_event;
#else
  unsigned short o_event;
#endif
#endif

  short o_tt, o_To;

#ifdef HAS_UNLESS
  short e_state;	/* if escape trans - state of origin */
#endif

#if defined(FULLSTACK) && !defined(MA)
  struct H_el *ostate;	/* pointer to stored state */
#endif

#ifdef CNTRSTACK
  long	j6, j7;
#endif

  char rt;              /* returning from function flag */
  char vt;              /* virtual transition flag */

#ifndef NOGC
  uchar gc;              /* garbagge collection counter */
#ifdef GC0
  uchar sv;              /* state saves counter */
#endif /* GC0 */
#endif /* NOGC */

  Trans *o_t;
  uint oval;
} Trail;

Trail *trail, *trpt;

short vt[MAXPROC + 1];
short rt[MAXPROC + 1];


/*** Some definitions ***/

int _; /* a predefined write-only variable */

long J1, J2, j1, j2, j3, j4;
long A_depth = 0, depth = 0, depth_adjust = 0;

double nstates = 0, truncs = 0;
int    mreached = 0, done = 0, errors = 0;
double nlinks = 0, nlost = 0, hcmp = 0;
int    single = 0, no_rck = 0;
int    mask, a_cycles = 0, upto = 1, strict = 0;
int    Nrun = 1, spsec = 0;

struct timeval start_tp, stop_tp;

double	truncs2 = 0;
double	ngrabs = 0;
unsigned long nr_states = 0; /* nodes in DFA */
long	Fa = 0, Fh = 0, Zh = 0, Zn = 0;
long	Ccheck = 0, Cholds = 0;

int     mask;

uchar	HASH_NR = 0;
unsigned int HASH_CONST[] = {
  /* asuming 4 bytes per int */
  0x88888EEF,	0x00400007,
  0x04c11db7,	0x100d4e63,
  0x0fc22f87,	0x3ff0c3ff,
  0x38e84cd7,	0x02b148e9,
  0x98b2e49d,	0xb616d379,
  0xa5247fd9,	0xbae92a15,
  0xb91c8bc5,	0x8e5880f3,
  0xacd7c069,	0xb4c44bb3,
  0x2ead1fb7,	0x8e428171,
  0xdbebd459,	0x828ae611,
  0x6cb25933,	0x86cdd651,
  0x9e8f5f21,	0xd5f8d8e7,
  0x9c4e956f,	0xb5cf2c71,
  0x2e805a6d,	0x33fc3a55,
  0xaf203ed1,	0xe31f5909,
  0x5276db35,	0x0c565ef7,
  0x273d1aa5,	0x8923b1dd,
  0
};

static double memcnt = 0;

/* loop detection */
int depthfound = -1;

/* blocked_on_queue status */
short boq = -1;

/* a predefined write-only variable */
int _;

long maxdepth = 10000;

#ifndef INLINE
int TstOnly=0;
#endif

#ifdef BITSTATE
int ssize=22;
#else
int ssize=18;
#endif

uchar warned = 0, iterative = 0, every_error = 0;
uchar noasserts = 0, noends = 0;

long PUT = 0, PROBE = 0, ZAPS = 0;
int state_tables=0, fairness=0, Nr_Trails=0;


#ifndef XUSAFE
char *q_name[MAXQ+1];
char *p_name[MAXPROC+1];
#endif /* XUSAFE */


uchar *SS, *LL;


/*** Exported function prototypes ***/

void  * emalloc(unsigned);
void  * erealloc(void *, unsigned int, int);

inline void push(int, int);
inline void pop(int, int);

inline uint   mkptr(uint, uchar, uchar);
inline uchar *deptr(uint, long);
inline int    del_index(uint);

inline char *xptr(char *);

inline void call(int, int, int);
inline void ret(int, int, int);

inline void cptr(int, uint);
inline uint fptr(int, int);
inline int  fpot(uint);

inline void poll(int, int, int);
inline void polp(int, uint);

inline int  h_index(uint);
inline void gc_inc(uint);
inline void gc_inc_back(uint);
inline void gc_dec(uint, int);
inline void gc_dec_back(uint);
inline void gc_scan(void);
inline void gc_sweep(int);

#ifdef POSYMM
void symm_get_scalar(scalar_t *, int);
#endif /* POSYMM */

#ifdef FULLSYMM
inline int  copy_perm(int *, int);
inline void repnow(int);
void        perm_heap(int *, int, int);
#endif /* FULLSYMM */

int     endstate(void);
void    assert(int, char *, int, int, Trans *);

void    r_ck(uchar *, int, int, short *);
void    putrail(void);
void    usage(FILE *);

void    usage(FILE *);
void    Uerror(char *);
void    uerror(char *);
void    fatal(char *);
void    wrap_stats(void);
void    wrapup(int);
void    snapshot(void);

void    _hinit(void);
void    _double_hash(State *);

#if defined(FULLSTACK) && defined(BITSTATE)
void    _onstack_init(void);
void    _onstack_zap(void);
void    _onstack_put(void);
int     _onstack_now(void);
#endif 

int     _statecmp(State *, State *);
State * _statedup(State *);
long    _statesz(State *);
void    _statedump(State *);
void    _statecpy(State *, State *);

int     _addproc(int);

#ifdef POSYMM
int     _addq(int, int, uchar, int, int, scalar_t *);
#else
int     _addq(int, int, uchar, int, int);
#endif /* POSYMM */

void    _delproc(int, int);
void    _delq(int, int, int);
void    _delheap(int);

#ifdef POSYMM
int     _addheap(int, uint, scalar_t *);
#else
int     _addheap(int, uint);
#endif /* POSYMM */

void    _sv_save(State *);
void    _sv_restor(void);

void    _p_restor(int);
void    _q_restor(int);

void    _p_save(int, int, uchar *);
void    _q_save(int);

void    _p_q_save(int, int);
void    _p_q_restor(int, int);

int     _hstore(State *);


#ifdef MA
int     _gstore(State *, uchar);
void	dfa_init(ushort);	
int	dfa_member(ushort);
int	dfa_store(uchar *);
void	dfa_stats(void);
void    r_xpoint();
#endif /* MA */


/** Other function prototypes **/

int     q_cond(short, Trans *);
int     q_full(int);
int     q_len(int);
int     q_zero(int);
int     qrecv(int, int, int, int);
void    unrecv(int, int, int, int, int);
int     unsend(int);

void    checkcycles(void);
void    crack(int, int, Trans *, short *);

void    imed(Trans *, int, int);
void    new_state(void);
void    putpeg(int, int);

void    retrans(int, int, int, short *, uchar *);
void    settable(void);
void    setq_claim(int, int, char *, int, char *);
void    tagtable(int, int, int, short *, uchar *);

void    struct_copy(char *, char *, long);

#ifndef NOBOUNDCHECK
int     Boundcheck(int, int, int, int, Trans *);
#endif

#ifndef XUSAFE
int     q_S_check(int, int);
int     q_R_check(int, int);
uchar   q_claim[MAXQ+1];
#endif /* XUSAFE */

#endif /* __SPIN_H__ */

