/***** spin: pangen4.c *****/

/* Copyright (c) 1991-1997 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#include "spin.h"
#ifdef PC
#include "y_tab.h"
#else
#include "y.tab.h"
#endif

extern FILE	*tc, *tb;
extern ProcList *Proc;
extern Queue	*qtab;
extern Symbol	*Fname, *context;
extern int	lineno, m_loss, Pid, eventmapnr;
extern short	nocast, has_provided, has_sorted;
extern char	*R13[], *R14[], *R15[];

static void	check_proc(Lextok *, int);


void
undostmnt(Lextok *now, int m)
{	
  Lextok *v;
  int i, j;

  if (!now)
    {	
      fprintf(tb, "0");
      return;
    }
	
  lineno = now->ln;
  Fname  = now->fn;

  switch (now->ntyp) 
    {
    case CONST:	case '!':	case UMIN:
    case '~':	case '/':	case '*':
    case '-':	case '+':	case '%':
    case LT:	case GT:	case '&':
    case '|':	case LE:	case GE:
    case NE:	case EQ:	case OR:
    case AND:	case LSHIFT:	case RSHIFT:
    case TIMEOUT:	case LEN:	case NAME:
    case LVAL:  case __NULL:
    case FULL:	case EMPTY:	case 'R':
    case NFULL:	case NEMPTY:	case ENABLED:
    case '?':	case PC_VAL:	case '^':
    case NONPROGRESS:
      putstmnt(tb, now, m);
      break;

    case RUN:
      fprintf(tb, "delproc(0, now._nr_pr-1)");
      break;

    case NEW:
    case DELETE:
      fprintf(tb, "sv_restor()");
      break;

    case RETFUN:
    case FUNCALL:
      fprintf(tb, "if (!rt[II]) {\n\t\t");
      fprintf(tb, "  sv_restor();\n\t\t");
      fprintf(tb, "  depth_adjust --;\n\t\t");
      fprintf(tb, "}\n\t\t");

      if (now->lft)
	{
	  fprintf(tb, "else {\n\t\t");

	  if (node_type2(now->lft) & POINTER)
	    {
	      fprintf(tb, "\n#ifdef GC0\n\t\t");
	      fprintf(tb, "gc_dec_back((uint) trpt->oval);\n");
	      fprintf(tb, "#endif\n\t\t");
	    }

	  putname(tb, "  ", now->lft, m, " = trpt->oval;\n\t\t", 0);
	  fprintf(tb, "}\n");
	}

      break;

    case RETURN:
      fprintf(tb, "sv_restor();\n\t\t");
      break;

    case 's':
      if (Pid == eventmapnr) break;
      
      if (m_loss)
	{	
	  fprintf(tb, "if (m == 2) m = unsend");
	  putname(tb, "(", now->lft, m, ")", 1);
	} 
      else {	
	fprintf(tb, "m = unsend");
	putname(tb, "(", now->lft, m, ")", 1);
      }
      
      break;

    case 'r':
      if (Pid == eventmapnr) break;

      for (v = now->rgt, i=j=0; v; v = v->rgt, i++)
	if (v->lft->ntyp != CONST
	    &&  v->lft->ntyp != EVAL)
	  j++;
		
      fprintf(tb, ";\n");
      if (j == 0 && now->val >= 2)
	{	
	  fprintf(tb, "\t\t");
	  break;	/* poll without side-effect */
	}
	
      if (j == 0 && now->val == 0)
	{	
	  fprintf(tb, "\t\tif (q_flds[qptr(");
	  putname(tb, "", now->lft, m, "-1)->_t]", 1);
	  fprintf(tb, " != %d)\n", i);
	} else
	  fprintf(tb, "\t\tif (1)\n");
      
      /*
       * fprintf(tb, "#if defined(FULLSTACK)");
       * fprintf(tb, " && defined(NOCOMP)");
       * fprintf(tb, " && !defined(BITSTATE)");
       * fprintf(tb, " && !defined(MA)\n");
       *
       * fprintf(tb, "#if VECTORSZ<=1024\n");
       * fprintf(tb, "\t\t\tsv_restor(!(t->atom&2));\n");
       * fprintf(tb, "#else\n");
       */

      fprintf(tb, "\t\t\tp_q_restor(II, ");
      putname(tb, "", now->lft, m, ");\n", 1);

      /*
       * fprintf(tb, "#endif\n");
       * fprintf(tb, "#else\n");
       *
       * fprintf(tb, "#if VECTORSZ<=1024\n");
       * fprintf(tb, "\t\t\tsv_restor(0);\n");
       * fprintf(tb, "#else\n");
       *
       * fprintf(tb, "\t\t\tp_q_restor(II, ");
       * putname(tb, "", now->lft, m, ");\n", 1);
       *
       * fprintf(tb, "#endif\n");
       * fprintf(tb, "#endif\n\t\t");
       */
	
      if (j == 0 && now->val == 0)
	{	
	  fprintf(tb, "else {\n\t\t");
	  for (v = now->rgt, i = 0; v; v = v->rgt, i++)
	    {	
	      fprintf(tb, "\tunrecv");
	      putname(tb, "(", now->lft, m, ", 0, ", 1);
	      fprintf(tb, "%d, ", i);
	      
	      if (v->lft->ntyp == EVAL)
		undostmnt(v->lft->lft, m);
	      else
		undostmnt(v->lft, m);
	      
	      fprintf(tb, ", %d);\n\t\t", (i==0)?1:0);
	    }
	  
	  fprintf(tb, "}\n\t\t");
	}
      
      break;

    case ASGN:
      {
	int lt, rt;

	if (!now->lft->sym)
	  fatal("no left value - undostmnt (ASGN)", (char *)0);

	lt = node_type2(now->lft);
	rt = node_type2(now->rgt);

	if (now->rgt->ntyp == NAME && (lt & POINTER) && (rt & POINTER))
	  {
	    fprintf(tb, "\n#ifdef GC0\n\t\t");
	    putname(tb, "gc_inc_back(", now->lft, m, ");\n", 0);
	    fprintf(tb, "#endif /* GC0 */\n");

	    fprintf(tb, "#ifdef GC1\n");
	    fprintf(tb, "\t\tif (trpt->gc)\n");
	    fprintf(tb, "\t\t  sv_restor();\n");
	    fprintf(tb, "#endif /* GC1 */\n\t\t");
	  }

	if (lt == BIT)
	  putname(tb, "", now->lft, m, "trpt->oval)", 0);
	else 
	  {
	    if ((lt & POINTER) && (rt & POINTER))
	      putname(tb, "", now->lft, m, " = trpt->oval", 0);
	    else
	      putname(tb, "", now->lft, m, " = trpt->oval", 1);
	  }

	fprintf(tb, ";\n\t\t");

	if ((lt & POINTER) && (rt & POINTER))
	  {
	    fprintf(tb, "\n#ifdef GC0\n\t\t");
	    putname(tb, "gc_dec_back(", now->lft, m, ");\n", 0);
	    fprintf(tb, "#endif\n\t\t");
	  }
	
	check_proc(now->rgt, m);
      }

      break;
      
    case 'c':
      check_proc(now->lft, m);
      break;
      
    case '.':
    case GOTO:
    case ELSE:
    case BREAK:
      break;
      
    case ASSERT:
    case PRINT:
      check_proc(now, m);
      break;

    case STOP: /* Stop */
      fprintf(tb, "p_restor(II);");

      fprintf(tb, "\n#ifdef  GC0\n");
      doformal(tb, GC0_DEC_UNDO, Proc->s, 0, 0);
      dolocal(tb, GC0_DEC_UNDO, Proc->s, 0);
      fprintf(tb, "#endif /* GC0 */\n");

      fprintf(tb, "#ifdef GC1\n\t\t");
      fprintf(tb, "if (trpt->gc) sv_restor();\n");
      fprintf(tb, "#endif /* GC1 */\n\t\t");

      break;
      
    default:
      printf("spin: bad node type %d `%c' (.b)\n", now->ntyp, now->ntyp);
      alldone(1);
    }
}


int
any_undo(Lextok *now)
{	/* is there anything to undo on a return move? */
	if (!now) return 1;
	switch (now->ntyp) {
	case 'c':	return any_oper(now->lft, RUN);
	case ASSERT:
	case PRINT:	return any_oper(now, RUN);

	case   '.':
	case  GOTO:
	case  ELSE:
	case BREAK:	return 0;
	default:	return 1;
	}
}


int
any_oper(Lextok *now, int oper)
{	/* check if an expression contains oper operator */
	if (!now) return 0;
	if (now->ntyp == oper)
		return 1;
	return (any_oper(now->lft, oper) || any_oper(now->rgt, oper));
}

static void
check_proc(Lextok *now, int m)
{
	if (!now)
		return;
	if (now->ntyp == STOP || now->ntyp == RUN)
	{	fprintf(tb, ";\n\t\t");
		undostmnt(now, m);
	}
	check_proc(now->lft, m);
	check_proc(now->rgt, m);
}


void
genunio(void)
{	
  char buf1[256];
  Queue *q; int i;

  ntimes(tc, 0, 1, R13);
  for (q = qtab; q; q = q->nxt)
    {	
      fprintf(tc,   "  case %d:\n", q->qid);

      if (has_sorted) {	
	sprintf(buf1, "((Q%d *z)->contents", q->qid);
	fprintf(tc, "    j = trpt->oval;\n\n");
	fprintf(tc, "    for (k = j; k < ((Q%d *)z)->Qlen; k ++) {\n", q->qid);
	
	for (i = 0; i < q->nflds; i ++)
	  fprintf(tc, "    %s[k].fld%d = %s[k+1].fld%d;\n", buf1, i, buf1, i);

	fprintf(tc, "    }\n\n");
	fprintf(tc, "    j = z->Qlen;\n");
      }

      sprintf(buf1, "((Q%d *)z)->contents[j].fld", q->qid);

      for (i = 0; i < q->nflds; i ++)
	fprintf(tc, "    %s%d = 0;\n", buf1, i);

      if (q->nslots == 0) {
	/* check if rendezvous succeeded, 1 level down */
	fprintf(tc, "    m = (trpt + 1)->o_m;\n");
	fprintf(tc, "    if (m) (trpt - 1)->o_pm |= 1;\n");
	fprintf(tc, "    UnBlock;\n");
      } else
	fprintf(tc, "    m = trpt->o_m;\n");

      fprintf(tc,   "    break;\n\n");
    }
	
  ntimes(tc, 0, 1, R14);

  for (q = qtab; q; q = q->nxt)
    {	
      sprintf(buf1,     "((Q%d *)z)->contents", q->qid);
      fprintf(tc,       "  case %d:\n", q->qid);

      if (q->nslots == 0)
	fprintf(tc,     "    if (strt) boq = from + 1;\n");
      else 
	if (q->nslots > 1) {	/* shift */
	  fprintf(tc,   "    if (strt && slot < %d) {\n", q->nslots - 1);
	  fprintf(tc,   "      for (j --; j >= slot; j --) {\n");

	  for (i = 0; i < q->nflds; i ++)
	    fprintf(tc, "        %s[j + 1].fld%d = %s[j].fld%d;\n", 
		    buf1, i, buf1, i);

	  fprintf(tc,  "        }\n      }\n");
	}

      strcat(buf1, "[slot].fld");
      fprintf(tc, "    if (strt) {\n");

      for (i = 0; i < q->nflds; i ++)
	fprintf(tc,    "        %s%d = 0;\n", buf1, i);

      fprintf(tc,      "    }\n");

      if (q->nflds == 1)	/* set */
	fprintf(tc,    "    if (fld == 0) %s0 = fldvar;\n", buf1);
      else {	
	fprintf(tc,    "    switch (fld) {\n");
	for (i = 0; i < q->nflds; i ++) {
	  fprintf(tc,  "    case %d:  %s%d = fldvar;\n", i, buf1, i);
	  fprintf(tc,  "      break;\n\n");
	}

	fprintf(tc,    "    }\n");
      }

      fprintf(tc,      "    break;\n\n");
    }
	
  ntimes(tc, 0, 1, R15);
}


int
proper_enabler(Lextok *n)
{
	if (!n) return 1;
	switch (n->ntyp) {
	case NEMPTY:	case FULL:
	case NFULL:	case EMPTY:
	case LEN:	case 'R':
	case NAME:
		has_provided = 1;
		if (strcmp(n->sym->name, "_pid") == 0)
			return 1;
		return (!(n->sym->context));

	case CONST:	case TIMEOUT:
		has_provided = 1;
		return 1;

	case ENABLED:	case PC_VAL:
		return proper_enabler(n->lft);

	case '!': case UMIN: case '~':
		return proper_enabler(n->lft);

	case '/': case '*': case '-': case '+':
	case '%': case LT:  case GT: case '&': case '^':
	case '|': case LE:  case GE:  case NE: case '?':
	case EQ:  case OR:  case AND: case LSHIFT:
	case RSHIFT: case 'c':
		return proper_enabler(n->lft) && proper_enabler(n->rgt);
	default:
		break;
	}
	return 0;
}
